/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * bltHash.h --
 *
 * Copyright 2015 George A. Howlett. All rights reserved.  
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are
 *   met:
 *
 *   1) Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2) Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the
 *      distribution.
 *   3) Neither the name of the authors nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 *   4) Products derived from this software may not be called "BLT" nor may
 *      "BLT" appear in their names without specific prior written
 *      permission from the author.
 *
 *   THIS SOFTWARE IS PROVIDED ''AS IS'' AND ANY EXPRESS OR IMPLIED
 *   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The hash table data structures are based upon the ones in the public header
 * file tcl.h in the Tcl library distribution.
 *
 *	Copyright (c) 1991-1993 The Regents of the University of California.
 *
 *	Copyright (c) 1994 Sun Microsystems, Inc.
 *
 *	See the file "license.terms" for information on usage and
 *	redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef BLT_HASH_H
#  define BLT_HASH_H 1

#  ifndef BLT_INT_H
#    ifndef SIZEOF_LONG
#      define SIZEOF_LONG 4
#    endif
#  ifndef SIZEOF_LONG_LONG
#    define SIZEOF_LONG_LONG 8
#  endif
#  ifndef SIZEOF_INT
#    define SIZEOF_INT 4
#  endif
#  ifndef SIZEOF_VOID_P
#    define SIZEOF_VOID_P 4
#  endif
#endif /* !BLT_INT_H */

#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#else
#  if (SIZEOF_VOID_P == 8)
#    if (SIZEOF_LONG == 8)
typedef signed long int64_t;
typedef unsigned long uint64_t;
#    else
typedef signed long long int64_t;
typedef unsigned long long uint64_t;
#    endif /* SIZEOF_LONG == 8 */
#  else
#    ifndef __CYGWIN__
typedef signed int int32_t;
typedef unsigned int uint32_t;
#    endif /* __CYGWIN__ */
#  endif /* SIZEOF_VOID_P == 8 */
#endif /* HAVE_INTTYPES_H */

#if (SIZEOF_VOID_P == 8) 
typedef uint64_t Blt_Hash;
#else
typedef uint32_t Blt_Hash;
#endif /* SIZEOF_VOID_P == 8 */

#include "bltPool.h"

/*
 * Acceptable key types for hash tables:
 */
#ifndef BLT_STRING_KEYS
#  define BLT_STRING_KEYS		0L
#endif
#ifndef BLT_ONE_WORD_KEYS
#  define BLT_ONE_WORD_KEYS	((size_t)-1L)
#endif
/*
 * Forward declaration of Blt_HashTable.  Needed by some C++ compilers to
 * prevent errors when the forward reference to Blt_HashTable is encountered
 * in the Blt_HashEntry structure.
 */

#ifdef __cplusplus
struct Blt_HashTable;
#endif

typedef union {			/* Key has one of these forms: */
    void *oneWordValue;		/* One-word value for key. */
#  if (SIZEOF_VOID_P == 8)
    uint64_t words[1];          /* Multiple integer words for key.  The
				 * actual size will be as large as
				 * necessary for this table's keys. */
#else
    uint32_t words[1];          /* Multiple integer words for key.  The
				 * actual size will be as large as
				 * necessary for this table's keys. */
#endif    
    char string[4];		/* String for key.  The actual size will be as
				 * large as needed to hold the key. */
} Blt_HashKey;

/*
 * Structure definition for an entry in a hash table.  No one outside BLT
 * should access any of these fields directly; use the macros defined below.
 */
typedef struct Blt_HashEntry {
    struct Blt_HashEntry *nextPtr; /* Pointer to next entry in this hash
				    * bucket, or NULL for end of chain.*/
    Blt_Hash hval;

    ClientData clientData;	/* Application stores something here with
				 * Blt_SetHashValue. */
    Blt_HashKey key;		/* MUST BE LAST FIELD IN RECORD!! */
} Blt_HashEntry;

/*
 * Structure definition for a hash table.  Must be in blt.h so clients can
 * allocate space for these structures, but clients should never access any
 * fields in this structure.
 */
#define BLT_SMALL_HASH_TABLE 4
typedef struct Blt_HashTable {
    Blt_HashEntry **buckets;	/* Pointer to bucket array.  Each element
				 * points to first entry in bucket's hash
				 * chain, or NULL. */
    Blt_HashEntry *staticBuckets[BLT_SMALL_HASH_TABLE];
				/* Bucket array used for small tables (to
				 * avoid mallocs and frees). */
    size_t numBuckets;		/* Total number of buckets allocated at
				 * buckets. */
    size_t numEntries;		/* Total # of entries present in table. */
    size_t rebuildSize;		/* Enlarge table when numEntries gets to be
				 * this large. */
    Blt_Hash mask;		/* Mask value used in hashing function. */	
    unsigned int downShift;	/* Shift count used in hashing function.
				 * Designed to use high- order bits of
				 * randomized keys. */
    size_t keyType;		/* Type of keys used in this table.  It's
				 * either BLT_STRING_KEYS, BLT_ONE_WORD_KEYS,
				 * or an integer giving the number of ints
				 * that is the size of the key. */
    Blt_HashEntry *(*findProc)(struct Blt_HashTable *tablePtr, const void *key);
    Blt_HashEntry *(*createProc)(struct Blt_HashTable *tablePtr, 
	const void *key, int *newPtr);

    Blt_Pool hPool;		/* Pointer to the pool allocator used for
				 * entries in this hash table. If NULL, the
				 * standard Tcl_Alloc, Tcl_Free routines will
				 * be used instead. */
} Blt_HashTable;

/*
 * Structure definition for information used to keep track of searches through
 * hash tables:
 */

typedef struct {
    Blt_HashTable *tablePtr;	/* Table being searched. */
#  if (SIZEOF_VOID_P == 8)
    uint64_t nextIndex;         /* Index of next bucket to be enumerated
				 * after present one. */
#else
    uint32_t nextIndex;         /* Index of next bucket to be enumerated
				 * after present one. */
#endif    
    Blt_HashEntry *nextEntryPtr; /* Next entry to be enumerated in the current
				  * bucket. */
} Blt_HashSearch;

/*
 * Macros for clients to use to access fields of hash entries:
 */

#define Blt_GetHashValue(h) ((h)->clientData)
#define Blt_SetHashValue(h, value) ((h)->clientData = (ClientData)(value))
#define Blt_GetHashKey(tablePtr, h) \
    ((void *) (((tablePtr)->keyType == BLT_ONE_WORD_KEYS) ? \
     (void *)(h)->key.oneWordValue : (h)->key.string))

/*
 * Macros to use for clients to use to invoke find and create procedures for
 * hash tables:
 */
#define Blt_FindHashEntry(tablePtr, key) \
    (*((tablePtr)->findProc))(tablePtr, (const char *)key)
#define Blt_CreateHashEntry(tablePtr, key, newPtr) \
    (*((tablePtr)->createProc))(tablePtr, (const char *)key, newPtr)

BLT_EXTERN void Blt_InitHashTable(Blt_HashTable *tablePtr, size_t keyType);

BLT_EXTERN void Blt_InitHashTableWithPool(Blt_HashTable *tablePtr, 
	size_t keyType);

BLT_EXTERN void Blt_DeleteHashTable(Blt_HashTable *tablePtr);

BLT_EXTERN void Blt_DeleteHashEntry(Blt_HashTable *tablePtr, 
	Blt_HashEntry *entryPtr);

BLT_EXTERN Blt_HashEntry *Blt_FirstHashEntry(Blt_HashTable *tablePtr, 
	Blt_HashSearch *searchPtr);

BLT_EXTERN Blt_HashEntry *Blt_NextHashEntry(Blt_HashSearch *srchPtr);

BLT_EXTERN const char *Blt_HashStats(Blt_HashTable *tablePtr);

#endif /* BLT_HASH_H */
